from datetime import datetime, timedelta, date
from typing import Self, Generator

from dvrd_pydate.enums import DatePart, TimePart
from dvrd_pydate.pydate import PyDate, CommonArg

hours_in_day = 24
minutes_in_hour = 60
seconds_in_minute = 60
microseconds_in_second = 1000


class PyDateTime(datetime, PyDate):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1:
            arg = args[0]
            if arg is None:
                return PyDateTime.now()
            if isinstance(arg, str):
                return PyDateTime.fromisoformat(arg)
            elif isinstance(arg, datetime):
                return datetime.__new__(cls, arg.year, arg.month, arg.day, arg.hour, arg.minute, arg.second,
                                        arg.microsecond, arg.tzinfo, fold=arg.fold)
            elif isinstance(arg, date):
                return datetime.__new__(cls, arg.year, arg.month, arg.day)
            elif isinstance(arg, int):
                return datetime.fromtimestamp(arg)
        elif len(args) == 2:
            arg_1 = args[0]
            arg_2 = args[1]
            if isinstance(arg_1, str) and isinstance(arg_2, str):
                return PyDateTime.strptime(arg_1, arg_2)
        if not args and not kwargs:
            now = datetime.now()
            return datetime.__new__(cls, now.year, now.month, now.day, now.hour, now.minute, now.second,
                                    now.microsecond, now.tzinfo, fold=now.fold)
        return datetime.__new__(cls, *args, **kwargs)

    @staticmethod
    def from_value(value: datetime | date | str | int = None) -> "PyDateTime":
        return PyDateTime(value)

    @staticmethod
    def iter(*, start: date | str = None, end: date | str | None = None,
             step: DatePart | TimePart | tuple[int | float, DatePart | TimePart] = DatePart.DAY,
             max_steps: int = None) -> \
            Generator["PyDateTime", None, None]:
        if max_steps == 0:
            # Raises StopIteration
            return
        if start is None:
            start = datetime.now()
        current = PyDateTime.from_value(start)
        end_value = None if end is None else PyDateTime.from_value(end)
        if isinstance(step, tuple):
            step_value = step[0]
            step_key = step[1]
        else:
            step_value = 1
            step_key = step
        current_step = 0
        while end_value is None or current < end_value:
            yield current
            current_step += 1
            if max_steps is not None and current_step == max_steps:
                break
            current = current.add(step_value, step_key)

    def set(self, value_or_key: CommonArg, key_or_value: CommonArg) -> Self:
        key, value = _determine_key_and_value(value_or_key, key_or_value)
        if isinstance(key, DatePart):
            return super().set(value, key)
        match key:
            case TimePart.HOURS | TimePart.HOUR:
                return self.set_hour(value)
            case TimePart.MINUTES | TimePart.MINUTE:
                return self.set_minute(value)
            case TimePart.SECONDS | TimePart.SECOND:
                return self.set_second(value)
            case TimePart.MICROSECONDS | TimePart.MICROSECOND:
                return self.set_microsecond(value)

    def add(self, value_or_key: CommonArg, key_or_value: CommonArg) -> Self:
        key, value = _determine_key_and_value(value_or_key, key_or_value)
        match key:
            case key if isinstance(key, DatePart):
                return super().add(value, key)
            case TimePart.HOURS | TimePart.HOUR:
                return self.add_hours(value)
            case TimePart.MINUTES | TimePart.MINUTE:
                return self.add_minutes(value)
            case TimePart.SECONDS | TimePart.SECOND:
                return self.add_seconds(value)
            case TimePart.MICROSECONDS | TimePart.MICROSECOND:
                return self.add_microseconds(value)

    def subtract(self, value_or_key: CommonArg, key_or_value: CommonArg) -> Self:
        key, value = _determine_key_and_value(value_or_key, key_or_value)
        match key:
            case key if isinstance(key, DatePart):
                return super().subtract(value, key)
            case TimePart.HOURS | TimePart.HOUR:
                return self.subtract_hours(value)
            case TimePart.MINUTES | TimePart.MINUTE:
                return self.subtract_minutes(value)
            case TimePart.SECONDS | TimePart.SECOND:
                return self.subtract_seconds(value)
            case TimePart.MICROSECONDS | TimePart.MICROSECOND:
                return self.subtract_microseconds(value)

    def set_hour(self, value: int) -> Self:
        return self.replace(hour=value)

    set_hours = set_hour

    def add_hours(self, value: int | float) -> Self:
        return self + timedelta(hours=value)

    def add_hour(self) -> Self:
        return self.add_hours(1)

    def subtract_hours(self, value: int | float) -> Self:
        return self - timedelta(hours=value)

    def subtract_hour(self) -> Self:
        return self.subtract_hours(1)

    def set_minute(self, value: int) -> Self:
        return self.replace(minute=value)

    set_minutes = set_minute

    def add_minutes(self, value: int | float) -> Self:
        return self + timedelta(minutes=value)

    def add_minute(self) -> Self:
        return self.add_minutes(1)

    def subtract_minutes(self, value: int | float) -> Self:
        return self - timedelta(minutes=value)

    def subtract_minute(self) -> Self:
        return self.subtract_minutes(1)

    def set_second(self, value: int) -> Self:
        return self.replace(second=value)

    set_seconds = set_second

    def add_seconds(self, value: int | float) -> Self:
        return self + timedelta(seconds=value)

    def add_second(self) -> Self:
        return self.add_seconds(1)

    def subtract_seconds(self, value: int | float) -> Self:
        return self - timedelta(seconds=value)

    def subtract_second(self) -> Self:
        return self.subtract_seconds(1)

    def set_microsecond(self, value: int) -> Self:
        return self.replace(microsecond=value)

    set_microseconds = set_microsecond

    def add_microseconds(self, value: int | float) -> Self:
        return self + timedelta(microseconds=value)

    def add_microsecond(self) -> Self:
        return self.add_microseconds(1)

    def subtract_microseconds(self, value: int | float) -> Self:
        return self - timedelta(microseconds=value)

    def subtract_microsecond(self) -> Self:
        return self.subtract_microseconds(1)

    def start_of(self, part: DatePart | TimePart) -> Self:
        if isinstance(part, DatePart):
            if part in [DatePart.DAY, DatePart.DAYS]:
                return self.replace(hour=0, minute=0, second=0, microsecond=0)
            return super().start_of(part)
        elif part in [TimePart.HOUR, TimePart.HOURS]:
            return self.replace(minute=0, second=0, microsecond=0)
        elif part in [TimePart.MINUTE, TimePart.MINUTES]:
            return self.replace(second=0, microsecond=0)
        elif part in [TimePart.SECOND, TimePart.SECONDS]:
            return self.replace(microsecond=0)
        elif part in [TimePart.MICROSECOND, TimePart.MICROSECONDS]:
            return self
        else:
            raise KeyError(f'Unsupported start_of part {part}')

    def end_of(self, part: DatePart | TimePart) -> Self:
        if isinstance(part, DatePart):
            if part in [DatePart.DAY, DatePart.DAYS]:
                return self.replace(hour=23, minute=59, second=59, microsecond=999)
            return super().end_of(part)
        elif part in [TimePart.HOUR, TimePart.HOURS]:
            return self.replace(minute=59, second=59, microsecond=999)
        elif part in [TimePart.MINUTE, TimePart.MINUTES]:
            return self.replace(second=59, microsecond=999)
        elif part in [TimePart.SECOND, TimePart.SECONDS]:
            return self.replace(microsecond=999)
        elif part in [TimePart.MICROSECOND, TimePart.MICROSECONDS]:
            return self
        else:
            raise KeyError(f'Unsupported end_of part {part}')

    def is_before(self, other: datetime | str, granularity: DatePart | TimePart = TimePart.MICROSECOND) -> bool:
        return super().is_before(other=other, granularity=granularity)

    def is_same_or_before(self, other: datetime | str, granularity: DatePart | TimePart = TimePart.MICROSECOND) -> bool:
        return super().is_same_or_before(other=other, granularity=granularity)

    def is_same(self, other: datetime | str, granularity: DatePart | TimePart = TimePart.MICROSECOND) -> bool:
        return super().is_same(other=other, granularity=granularity)

    def is_same_or_after(self, other: datetime | str, granularity: DatePart | TimePart = TimePart.MICROSECOND) -> bool:
        return super().is_same_or_after(other=other, granularity=granularity)

    def is_after(self, other: datetime | str, granularity: DatePart | TimePart = TimePart.MICROSECOND) -> bool:
        return super().is_after(other=other, granularity=granularity)

    def is_between(self, other1: date | str, other2: date | str, *,
                   granularity: DatePart | TimePart = TimePart.MICROSECOND,
                   from_inclusive: bool = True, to_inclusive: bool = True) -> bool:
        return super().is_between(other1=other1, other2=other2, granularity=granularity, from_inclusive=from_inclusive,
                                  to_inclusive=to_inclusive)

    def py_date(self) -> PyDate:
        return PyDate(self.year, self.month, self.day)


def _determine_key_and_value(arg1: CommonArg, arg2: CommonArg) -> tuple[DatePart | TimePart, int]:
    arg1 = _number_or_date_time_part(arg1)
    arg2 = _number_or_date_time_part(arg2)
    key = value = None
    if isinstance(arg1, (int, float)):
        value = arg1
    elif isinstance(arg1, (DatePart, TimePart)):
        key = arg1

    if isinstance(arg2, (int, float)):
        value = arg2
    elif isinstance(arg2, (DatePart, TimePart)):
        key = arg2

    if key is None or value is None:
        raise ValueError('Key and/or value cannot be None')
    return key, value


def _number_or_date_time_part(arg: CommonArg) -> int | float | DatePart | TimePart:
    if isinstance(arg, str):
        return DatePart.get_item(arg) or TimePart.get_item(arg)
    return arg



import unittest
from calendar import monthrange
from datetime import date, timedelta

from dvrd_pydate import PyDateTime
from dvrd_pydate.enums import DatePart, TimePart
from dvrd_pydate.pydate import PyDate


class TestPyDate(unittest.TestCase):
    def setUp(self):
        self.test_date = PyDate(2023, 1, 15)

    def test_constructors(self):
        self.assertEqual(date(2023, 1, 15), PyDate.from_value("2023-01-15"))
        self.assertEqual(date(2023, 1, 15), PyDate("2023-01-15"))
        self.assertEqual(date(2023, 1, 15), PyDate(2023, 1, 15))
        self.assertEqual(date(2023, 1, 15), PyDate(date(2023, 1, 15)))
        self.assertEqual(date(2023, 1, 15), PyDate.parse_date(value='15-01-2023', fmt='%d-%m-%Y'))
        self.assertEqual(date.today(), PyDate.from_value())
        self.assertEqual(date.today(), PyDate())
        self.assertEqual(date.today(), PyDate(None))

    def test_add_operations(self):
        test_date = PyDate(2023, 1, 15)

        # Test adding years
        date_copy = PyDate.from_value(test_date).add(1, DatePart.YEAR)
        self.assertEqual(date(2024, 1, 15), date_copy)

        # Test adding months
        date_copy = PyDate.from_value(test_date).add(2, DatePart.MONTHS)
        self.assertEqual(date_copy, date(2023, 3, 15))

        # Test adding weeks
        date_copy = PyDate.from_value(test_date).add(1, DatePart.WEEKS)
        self.assertEqual(date_copy, date(2023, 1, 22))

        # Test adding days
        date_copy = PyDate.from_value(test_date).add(5, DatePart.DAYS)
        self.assertEqual(date_copy, date(2023, 1, 20))

        self.assertRaises(TypeError, date_copy.add, 5, TimePart.HOURS)

    def test_add_overflow_operations(self):
        test_date = PyDate.from_value('2023-12-31')

        # Test adding years
        date_copy = test_date.clone().add(1, DatePart.YEAR)
        self.assertEqual(date_copy, date(2024, 12, 31))

        date_copy = test_date.clone().add_year()
        self.assertEqual(date_copy, date(2024, 12, 31))

        # Test adding months
        date_copy = test_date.clone().add(1, DatePart.MONTH)
        self.assertEqual(date_copy, date(2024, 1, 31))

        date_copy = test_date.clone().add_month()
        self.assertEqual(date_copy, date(2024, 1, 31))

        # Test adding month to less max days
        date_copy = PyDate.from_value('2023-03-31').add(1, DatePart.MONTH)
        self.assertEqual(date_copy, date(2023, 4, 30))

        date_copy = PyDate.from_value('2023-03-31').add_month()
        self.assertEqual(date_copy, date(2023, 4, 30))

        # Test adding weeks
        date_copy = test_date.clone().add(1, DatePart.WEEK)
        self.assertEqual(date_copy, date(2024, 1, 7))

        date_copy = test_date.clone().add_week()
        self.assertEqual(date_copy, date(2024, 1, 7))

        # Test adding days
        date_copy = test_date.clone().add(1, DatePart.DAY)
        self.assertEqual(date_copy, date(2024, 1, 1))

        date_copy = test_date.clone().add_day()
        self.assertEqual(date_copy, date(2024, 1, 1))

    def test_subtract_operations(self):
        # Test subtracting years
        test_date = PyDate(2023, 1, 15)

        date_copy = PyDate.from_value(test_date).subtract(1, DatePart.YEAR)
        self.assertEqual(date_copy, date(2022, 1, 15))

        date_copy = PyDate.from_value(test_date).subtract_year()
        self.assertEqual(date_copy, date(2022, 1, 15))

        # Test subtracting months
        date_copy = PyDate.from_value(test_date).subtract(2, DatePart.MONTHS)
        self.assertEqual(date(2022, 11, 15), date_copy)

        date_copy = PyDate.from_value(test_date).subtract_month()
        self.assertEqual(date_copy, date(2022, 12, 15))

        # Test subtracting 37 in months
        date_copy = PyDate.from_value(test_date).subtract(37, DatePart.MONTHS)
        self.assertEqual(date_copy, date(2019, 12, 15))

        # Test subtracting weeks
        date_copy = PyDate.from_value(test_date).subtract(1, DatePart.WEEK)
        self.assertEqual(date_copy, date(2023, 1, 8))

        date_copy = PyDate.from_value(test_date).subtract_week()
        self.assertEqual(date_copy, date(2023, 1, 8))

        # Test subtracting days
        date_copy = PyDate.from_value(test_date).subtract(1, DatePart.DAY)
        self.assertEqual(date(2023, 1, 14), date_copy)

        date_copy = PyDate.from_value(test_date).subtract_day()
        self.assertEqual(date(2023, 1, 14), date_copy)

        self.assertRaises(TypeError, date_copy.subtract, 5, TimePart.HOURS)

    def test_subtract_overflow_operations(self):
        test_date = PyDate.from_value('2023-01-31')

        # Test subtracting years
        date_copy = test_date.clone().subtract(1, DatePart.YEAR)
        self.assertEqual(date(2022, 1, 31), date_copy)

        # Test subtracting months
        date_copy = test_date.clone().subtract(1, DatePart.MONTH)
        self.assertEqual(date(2022, 12, 31), date_copy)

        # Test subtracting months to less max days
        date_copy = PyDate.from_value('2023-07-31').subtract(1, DatePart.MONTH)
        self.assertEqual(date(2023, 6, 30), date_copy)

    def test_clone(self):
        cloned = self.test_date.clone()
        self.assertEqual(cloned, self.test_date)
        self.assertIsNot(cloned, self.test_date)

    def test_iter(self):
        # Default iter, with end date
        end = PyDate.today().add(1, DatePart.MONTHS)
        expect_date = date.today()
        for value in PyDate.iter(end=end):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(days=1)

        # Specifying start
        start = date(2024, 1, 1)
        end = date(2024, 1, 31)
        expect_date = date(2024, 1, 1)
        for value in PyDate.iter(start=start, end=end):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(days=1)

        # 2-days interval
        start = date(2024, 1, 1)
        end = date(2024, 1, 31)
        expect_date = date(2024, 1, 1)
        for value in PyDate.iter(start=start, end=end, step=(2, DatePart.DAYS)):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(days=2)

        # 1-month interval
        start = date(2024, 1, 1)
        end = date(2025, 7, 31)
        expect_date = date(2024, 1, 1)
        last_value = None
        for value in PyDate.iter(start=start, end=end, step=DatePart.MONTH):
            self.assertEqual(expect_date, value)
            next_month = expect_date.month + 1
            add_year, month_value = divmod(next_month, 13)
            if add_year:
                month_value += 1
            expect_date = expect_date.replace(year=expect_date.year + add_year, month=month_value)
            last_value = value
        self.assertLess(last_value, end)

        # Max steps argument
        result = PyDate.iter(max_steps=5)
        self.assertEqual(len(list(result)), 5)
        self.assertRaises(StopIteration, lambda: next(PyDate.iter(max_steps=0)))

        # Max steps + end
        result = PyDate.iter(end=PyDate.today().add(4, DatePart.DAYS), max_steps=5)
        self.assertEqual(len(list(result)), 4)

        result = PyDate.iter(end=PyDate.today().add(6, DatePart.DAYS), max_steps=5)
        self.assertEqual(len(list(result)), 5)

        # Invalid interval
        self.assertRaises(KeyError,
                          lambda: next(PyDate.iter(start=PyDate.today(), end=PyDate.today(), step=TimePart.HOURS)))
        self.assertRaises(KeyError,
                          lambda: next(PyDate.iter(start=PyDate.today(), end=PyDate.today(), step=(1, TimePart.HOURS))))

    def test_start_of(self):
        test_date = PyDate.today()
        start_of = test_date.start_of(DatePart.YEAR)
        self.assertTupleEqual((start_of.month, start_of.day), (1, 1))

        start_of = test_date.start_of(DatePart.MONTH)
        self.assertTupleEqual((start_of.month, start_of.day), (date.today().month, 1))

        test_date = PyDate.fromisoformat('2024-11-24')  # Sunday
        self.assertEqual(date(2024, 11, 18), test_date.start_of(DatePart.WEEK))

        self.assertIs(test_date, test_date.start_of(DatePart.DAY))
        self.assertRaises(KeyError, test_date.start_of, TimePart.HOURS)
        self.assertRaises(KeyError, test_date.start_of, 'not_a_key')

    def test_end_of(self):
        test_date = PyDate.today()

        # Year
        end_of = test_date.end_of(DatePart.YEAR)
        self.assertTupleEqual((end_of.month, end_of.day), (12, 31))

        # Month
        end_of = test_date.end_of(DatePart.MONTH)
        today = date.today()
        max_date = monthrange(today.year, today.month)[1]
        self.assertTupleEqual((end_of.month, end_of.day), (today.month, max_date))

        # Week
        test_date = PyDate.fromisoformat('2024-11-20')  # Wednesday
        end_of = test_date.end_of(DatePart.WEEK)
        self.assertEqual(date(2024, 11, 24), end_of)

        self.assertIs(test_date, test_date.end_of(DatePart.DAY))
        self.assertRaises(KeyError, test_date.end_of, TimePart.HOURS)
        self.assertRaises(KeyError, test_date.end_of, 'not_a_key')

    def test_is_before(self):
        # Full date check
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 11, 25)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)

        self.assertTrue(date1.is_before(date2))
        self.assertTrue(date1.is_before(date2_value))
        self.assertFalse(date2.is_before(date1))
        self.assertFalse(date2.is_before(date1_value))

        # Month check
        date3 = PyDate(2024, 10, 24)
        self.assertFalse(date1.is_before(date2, DatePart.MONTH))
        self.assertFalse(date2.is_before(date1, DatePart.MONTH))

        self.assertTrue(date3.is_before(date1, DatePart.MONTH))
        self.assertTrue(date3.is_before(date2, DatePart.MONTH))

        # Year check
        date4 = PyDate(2023, 11, 24)
        self.assertFalse(date1.is_before(date2, DatePart.YEAR))
        self.assertFalse(date2.is_before(date1, DatePart.YEAR))
        self.assertFalse(date1.is_before(date3, DatePart.YEAR))
        self.assertFalse(date2.is_before(date1, DatePart.YEAR))
        self.assertFalse(date1.is_before(date4, DatePart.YEAR))

        self.assertTrue(date4.is_before(date1, DatePart.YEAR))
        self.assertTrue(date4.is_before(date2, DatePart.YEAR))
        self.assertTrue(date4.is_before(date3, DatePart.YEAR))

    def test_is_same_or_before(self):
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 11, 25)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)
        date3 = date2.clone()

        # Full date check
        self.assertTrue(date1.is_same_or_before(date2))
        self.assertTrue(date1.is_same_or_before(date2_value))
        self.assertTrue(date2.is_same_or_before(date3))
        self.assertTrue(date3.is_same_or_before(date2))

        self.assertFalse(date2.is_same_or_before(date1))
        self.assertFalse(date2.is_same_or_before(date1_value))
        self.assertFalse(date2.is_same_or_before(date1))
        self.assertFalse(date3.is_same_or_before(date1))

        # Month check
        date4 = PyDate(2023, 10, 24)
        self.assertTrue(date1.is_same_or_before(date2, DatePart.MONTH))
        self.assertTrue(date2.is_same_or_before(date1, DatePart.MONTH))
        self.assertTrue(date1.is_same_or_before(date3, DatePart.MONTH))
        self.assertTrue(date3.is_same_or_before(date1, DatePart.MONTH))

        self.assertFalse(date1.is_same_or_before(date4, DatePart.MONTH))

        # Day check
        self.assertTrue(date1.is_same_or_before(date2, DatePart.DAY))
        self.assertTrue(date2.is_same_or_before(date3, DatePart.DAY))
        self.assertTrue(date3.is_same_or_before(date2, DatePart.DAY))
        self.assertTrue(date4.is_same_or_before(date1, DatePart.DAY))
        self.assertTrue(date4.is_same_or_before(date2, DatePart.DAY))
        self.assertTrue(date4.is_same_or_before(date3, DatePart.DAY))

        self.assertFalse(date2.is_same_or_before(date1, DatePart.DAY))
        self.assertFalse(date3.is_same_or_before(date1, DatePart.DAY))

    def test_is_same(self):
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 11, 25)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)

        # Full check
        self.assertFalse(date1.is_same(date2))
        self.assertFalse(date1.is_same(date2_value))
        self.assertFalse(date2.is_same(date1))

        # Year
        self.assertTrue(date1.is_same(date2, DatePart.YEAR))
        self.assertTrue(date2.is_same(date1, DatePart.YEAR))

        # Month
        self.assertTrue(date1.is_same(date2, DatePart.MONTH))
        self.assertTrue(date2.is_same(date1, DatePart.MONTH))

        # Day
        self.assertFalse(date1.is_same(date2, DatePart.DAY))
        self.assertFalse(date2.is_same(date1, DatePart.DAY))

    def test_is_same_or_after(self):
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 11, 25)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)

        # Full check
        self.assertTrue(date1.is_same_or_after(date1.clone()))
        self.assertTrue(date1.is_same_or_after(date1_value))
        self.assertTrue(date2.is_same_or_after(date1))

        self.assertFalse(date1.is_same_or_after(date2))

        # Year check
        date3 = PyDate(2023, 11, 25)
        self.assertTrue(date1.is_same_or_after(date2, DatePart.YEAR))
        self.assertTrue(date2.is_same_or_after(date1, DatePart.YEAR))

        self.assertFalse(date3.is_same_or_after(date1, DatePart.YEAR))

        # Month check
        self.assertTrue(date1.is_same_or_after(date2, DatePart.MONTH))
        self.assertTrue(date2.is_same_or_after(date1, DatePart.MONTH))

        self.assertFalse(date3.is_same_or_after(date1, DatePart.MONTH))

        # Day check
        self.assertTrue(date1.is_same_or_after(date1.clone(), DatePart.DAY))
        self.assertTrue(date2.is_same_or_after(date1, DatePart.DAY))

        self.assertFalse(date1.is_same_or_after(date2, DatePart.DAY))

    def test_is_after(self):
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 11, 25)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)

        # Full check
        self.assertTrue(date2.is_after(date1))
        self.assertTrue(date2.is_after(date1_value))

        self.assertFalse(date1.is_after(date2))
        self.assertFalse(date1.is_after(date1.clone()))

        # Year
        date3 = PyDate(2023, 11, 24)
        self.assertTrue(date1.is_after(date3, DatePart.YEAR))

        self.assertFalse(date1.is_after(date2, DatePart.YEAR))
        self.assertFalse(date2.is_after(date1, DatePart.YEAR))

        # Month
        self.assertTrue(date1.is_after(date3, DatePart.MONTH))

        self.assertFalse(date1.is_after(date2, DatePart.MONTH))
        self.assertFalse(date2.is_after(date1, DatePart.MONTH))

        # Day
        self.assertTrue(date2.is_after(date1, DatePart.DAY))

        self.assertFalse(date1.is_after(date2, DatePart.DAY))
        self.assertFalse(date1.is_after(date1.clone(), DatePart.DAY))

    def test_is_between(self):
        date1_value = date(2024, 11, 24)
        date2_value = date(2024, 12, 24)
        date3_value = date(2024, 10, 24)
        date1 = PyDate.from_value(date1_value)
        date2 = PyDate.from_value(date2_value)
        date3 = PyDate.from_value(date3_value)

        # Intentionally give arguments in reverse order to make sure the determination of the `from_date`
        # and `to_date` works correctly
        self.assertTrue(date1.is_between(date2, date3))
        self.assertTrue(date1.is_between(date2_value, date3_value))
        self.assertFalse(date2.is_between(date1, date3))
        self.assertFalse(date3.is_between(date1, date2))

        # Test to date not inclusive
        date4 = date1.clone()
        self.assertTrue(date1.is_between(date3, date4, to_inclusive=True))
        self.assertFalse(date1.is_between(date3, date4, to_inclusive=False))

        # Test from date not inclusive
        self.assertTrue(date1.is_between(date4, date2, from_inclusive=True))
        self.assertFalse(date1.is_between(date4, date2, from_inclusive=False))

        # Granularity
        date2 = PyDate(2024, 11, 12)
        date3 = PyDate(2024, 11, 25)
        self.assertTrue(date1.is_between(date2, date3))
        self.assertTrue(date1.is_between(date3, date2, granularity=DatePart.MONTH))
        self.assertFalse(date1.is_between(date2, date3, granularity=DatePart.MONTH, from_inclusive=False))
        self.assertFalse(date1.is_between(date2, date3, granularity=DatePart.MONTH, to_inclusive=False))

        self.assertTrue(date1.is_between(date3, date2, granularity=DatePart.YEAR))
        self.assertFalse(date1.is_between(date2, date3, granularity=DatePart.YEAR, from_inclusive=False))
        self.assertFalse(date1.is_between(date2, date3, granularity=DatePart.YEAR, to_inclusive=False))

    def test_set_operations(self):
        pydate = PyDate(2024, 1, 1)
        self.assertEqual(date(2024, 2, 1), pydate.set(DatePart.MONTH, 2))
        self.assertEqual(date(2024, 2, 1), pydate.set(2, DatePart.MONTH))
        self.assertEqual(date(2024, 2, 1), pydate.set(2, 'month'))
        self.assertEqual(date(2024, 2, 1), pydate.set('months', 2))
        self.assertEqual(date(2024, 2, 1), pydate.set(2, 'months'))
        self.assertEqual(date(2024, 2, 1), pydate.set_month(2))
        self.assertEqual(date(2024, 2, 1), pydate.set_months(2))

        self.assertEqual(date(2024, 1, 2), pydate.set(DatePart.DAY, 2))
        self.assertEqual(date(2024, 1, 2), pydate.set(2, DatePart.DAYS))
        self.assertEqual(date(2024, 1, 2), pydate.set_day(2))
        self.assertEqual(date(2024, 1, 2), pydate.set_days(2))

        self.assertEqual(date(2025, 1, 1), pydate.set(DatePart.YEAR, 2025))
        self.assertEqual(date(2025, 1, 1), pydate.set(2025, DatePart.YEARS))
        self.assertEqual(date(2025, 1, 1), pydate.set_year(2025))
        self.assertEqual(date(2025, 1, 1), pydate.set_years(2025))

        self.assertRaises(TypeError, pydate.set, TimePart.HOURS, 2025)
        self.assertRaises(ValueError, pydate.set, None, 2025)
        self.assertRaises(ValueError, pydate.set, DatePart.DAY, None)

    def test_py_datetime(self):
        self.assertEqual(PyDateTime(2024, 1, 1, 0, 0, 0), PyDate(2024, 1, 1).py_datetime())


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime, timedelta, date

from dvrd_pydate import PyDate
from dvrd_pydate.enums import DatePart, TimePart
from dvrd_pydate.pydatetime import PyDateTime


class TestPyDateTime(unittest.TestCase):
    def setUp(self):
        self.test_date = datetime(2023, 1, 1, 12, 30, 45, 123456)
        self.py_datetime = PyDateTime.from_value(self.test_date)

    def test_initialization(self):
        self.assertEqual(self.test_date, PyDateTime.from_value(self.test_date))
        test_datetime = datetime.now()
        self.assertTrue((PyDateTime.from_value(test_datetime.date()) - test_datetime).total_seconds() < 1)

        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime.from_value('2023-01-01 00:00:00.000'))
        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime(2023, 1, 1, 0, 0, 0, 0))
        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime(datetime(2023, 1, 1, 0, 0, 0, 0)))
        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime(date(2023, 1, 1)))
        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime('2023-01-01 00:00:00.000'))
        self.assertEqual(datetime(2023, 1, 1, 0, 0, 0, 0), PyDateTime('2023-01-01 00:00:00', '%Y-%m-%d %H:%M:%S'))

        now = datetime.now()
        self.assertTrue((PyDateTime.from_value() - now).total_seconds() < 1)
        self.assertTrue((PyDateTime() - now).total_seconds() < 1)

    def test_add_methods(self):
        # Test adding hours
        result = self.py_datetime.clone().add(2, TimePart.HOURS)
        expected = self.test_date + timedelta(hours=2)
        self.assertEqual(expected, result)

        result = self.py_datetime.clone().add_hour()
        expected = self.test_date + timedelta(hours=1)
        self.assertEqual(expected, result)

        # Test adding minutes
        result = self.py_datetime.clone().add(30, TimePart.MINUTES)
        expected = self.test_date + timedelta(minutes=30)
        self.assertEqual(expected, result)

        result = self.py_datetime.clone().add_minute()
        expected = self.test_date + timedelta(minutes=1)
        self.assertEqual(expected, result)

        # Test adding seconds
        result = self.py_datetime.clone().add(30, TimePart.SECOND)
        expected = self.test_date + timedelta(seconds=30)
        self.assertEqual(expected, result)

        result = self.py_datetime.clone().add_second()
        expected = self.test_date + timedelta(seconds=1)
        self.assertEqual(expected, result)

        # Test adding microseconds
        result = self.py_datetime.clone().add(30, TimePart.MICROSECOND)
        expected = self.test_date + timedelta(microseconds=30)
        self.assertEqual(expected, result)

        result = self.py_datetime.clone().add_microsecond()
        expected = self.test_date + timedelta(microseconds=1)
        self.assertEqual(expected, result)

        self.assertRaises(ValueError, self.py_datetime.add, 30, 'not_a_part')

    def test_subtract_methods(self):
        # Test subtracting date part
        result = self.py_datetime.clone().subtract(1, DatePart.DAY)
        expected = self.test_date - timedelta(days=1)
        self.assertEqual(result, expected)

        # Test subtracting hours
        result = self.py_datetime.clone().subtract(2, TimePart.HOURS)
        expected = self.test_date - timedelta(hours=2)
        self.assertEqual(result, expected)

        result = self.py_datetime.clone().subtract_hour()
        expected = self.test_date - timedelta(hours=1)
        self.assertEqual(result, expected)

        # Test subtracting minutes
        result = self.py_datetime.clone().subtract(30, TimePart.MINUTES)
        expected = self.test_date - timedelta(minutes=30)
        self.assertEqual(result, expected)

        result = self.py_datetime.clone().subtract_minute()
        expected = self.test_date - timedelta(minutes=1)
        self.assertEqual(result, expected)

        # Test subtracting seconds
        result = self.py_datetime.clone().subtract(30, TimePart.SECOND)
        expected = self.test_date - timedelta(seconds=30)
        self.assertEqual(result, expected)

        result = self.py_datetime.clone().subtract_second()
        expected = self.test_date - timedelta(seconds=1)
        self.assertEqual(result, expected)

        # Test subtracting microseconds
        result = self.py_datetime.clone().subtract(30, TimePart.MICROSECOND)
        expected = self.test_date - timedelta(microseconds=30)
        self.assertEqual(result, expected)

        result = self.py_datetime.clone().subtract_microsecond()
        expected = self.test_date - timedelta(microseconds=1)
        self.assertEqual(result, expected)

        self.assertRaises(ValueError, self.py_datetime.subtract, 30, 'not_a_part')

    def test_clone(self):
        clone = self.py_datetime.clone()
        self.assertEqual(clone, self.py_datetime)
        self.assertIsNot(clone, self.py_datetime)

    def test_iter(self):
        # Default iter, with end date
        expect_date = datetime.now()
        end = PyDateTime.from_value(expect_date).add(1, DatePart.MONTHS)
        for value in PyDateTime.iter(end=end):
            self.assertTrue((value - expect_date).total_seconds() < 1)
            expect_date += timedelta(days=1)

        start = datetime(2024, 1, 1, 0, 0, 0, 0)
        end = datetime(2024, 1, 31, 0, 0, 0, 0)
        expect_date = datetime(2024, 1, 1, 0, 0, 0, 0)
        for value in PyDateTime.iter(start=start, end=end):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(days=1)

        start = datetime(2024, 1, 1, 0, 0, 0, 0)
        end = datetime(2024, 1, 31, 0, 0, 0, 0)
        expect_date = datetime(2024, 1, 1, 0, 0, 0, 0)
        for value in PyDateTime.iter(start=start, end=end, step=(2, DatePart.DAYS)):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(days=2)

        start = datetime(2024, 1, 1, 0, 0, 0, 0)
        end = datetime(2024, 1, 31, 0, 0, 0, 0)
        expect_date = datetime(2024, 1, 1, 0, 0, 0, 0)
        for value in PyDateTime.iter(start=start, end=end, step=(1, TimePart.HOUR)):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(hours=1)

        start = datetime(2024, 1, 1, 0, 0, 0, 0)
        end = datetime(2024, 1, 31, 0, 0, 0, 0)
        expect_date = datetime(2024, 1, 1, 0, 0, 0, 0)
        for value in PyDateTime.iter(start=start, end=end, step=(2, TimePart.MINUTE)):
            self.assertEqual(expect_date, value)
            expect_date += timedelta(minutes=2)

        result = PyDateTime.iter(max_steps=5)
        self.assertEqual(len(list(result)), 5)
        self.assertRaises(StopIteration, lambda: next(PyDateTime.iter(max_steps=0)))

    def test_start_of(self):
        now = datetime.now()
        start_of = PyDateTime.from_value(now)

        # Date part
        expected = now.replace(month=1, day=1)
        self.assertEqual(expected, start_of.start_of(DatePart.YEAR))

        # Day
        expected = datetime.combine(date.today(), datetime.min.time())
        self.assertEqual(expected, start_of.start_of(DatePart.DAY))

        # Hour
        expected = now.replace(minute=0, second=0, microsecond=0)
        self.assertEqual(expected, start_of.start_of(TimePart.HOUR))

        # Minute
        expected = now.replace(second=0, microsecond=0)
        self.assertEqual(expected, start_of.start_of(TimePart.MINUTE))

        # Second
        expected = now.replace(microsecond=0)
        self.assertEqual(expected, start_of.start_of(TimePart.SECOND))

        self.assertIs(start_of, start_of.start_of(TimePart.MICROSECONDS))
        self.assertRaises(KeyError, start_of.start_of, 'not_a_part')

    def test_end_of(self):
        now = datetime.now()
        end_of = PyDateTime.from_value(now)

        # Date part
        expected = now.replace(month=12, day=31)
        self.assertEqual(expected, end_of.end_of(DatePart.YEAR))

        # Day
        expected = now.replace(hour=23, minute=59, second=59, microsecond=999)
        self.assertEqual(expected, end_of.end_of(DatePart.DAY))

        # Hour
        expected = now.replace(minute=59, second=59, microsecond=999)
        self.assertEqual(expected, end_of.end_of(TimePart.HOUR))

        # Minute
        expected = now.replace(second=59, microsecond=999)
        self.assertEqual(expected, end_of.end_of(TimePart.MINUTE))

        # Second
        expected = now.replace(microsecond=999)
        self.assertEqual(expected, end_of.end_of(TimePart.SECOND))

        self.assertIs(end_of, end_of.end_of(TimePart.MICROSECONDS))
        self.assertRaises(KeyError, end_of.end_of, 'not_a_part')

    def test_set_operations(self):
        pydate = PyDateTime('2024-01-01 00:00:00')

        # Set date part
        self.assertEqual(datetime(2024, 1, 2, 0, 0, 0), pydate.set('day', 2))
        self.assertNotEqual(datetime(2024, 1, 1, 0, 0, 0), pydate.set('day', 2))

        # Set time part
        self.assertEqual(datetime(2024, 1, 1, 2, 0, 0), pydate.set('hour', 2))
        self.assertEqual(datetime(2024, 1, 1, 2, 0, 0), pydate.set(TimePart.HOUR, 2))
        self.assertEqual(datetime(2024, 1, 1, 2, 0, 0), pydate.set(2, 'hour'))
        self.assertEqual(datetime(2024, 1, 1, 2, 0, 0), pydate.set(2, TimePart.HOUR))

        self.assertEqual(datetime(2024, 1, 1, 0, 2, 0), pydate.set(TimePart.MINUTE, 2))
        self.assertEqual(datetime(2024, 1, 1, 0, 0, 2), pydate.set(TimePart.SECONDS, 2))
        self.assertEqual(datetime(2024, 1, 1, 0, 0, 0, 2), pydate.set(TimePart.MICROSECOND, 2))

    def test_py_date(self):
        self.assertEqual(PyDate(2024, 1, 1), PyDateTime(2024, 1, 1, 0, 0, 0).py_date())
        # Test that time doesn't matter here
        self.assertEqual(PyDate(2024, 1, 1), PyDateTime(2024, 1, 1, 23, 59, 59).py_date())


if __name__ == '__main__':
    unittest.main()



from enum import Enum


class BaseEnum(Enum):
    @classmethod
    def get_item(cls, value: str):
        return next((item for item in cls if item.value == value), None)


class DatePart(BaseEnum):
    YEAR = 'year'
    YEARS = 'years'
    MONTH = 'month'
    MONTHS = 'months'
    DAY = 'day'
    DAYS = 'days'
    WEEK = 'week'
    WEEKS = 'weeks'


class TimePart(BaseEnum):
    HOUR = 'hour'
    HOURS = 'hours'
    MINUTE = 'minute'
    MINUTES = 'minutes'
    SECOND = 'second'
    SECONDS = 'seconds'
    MICROSECOND = 'microsecond'
    MICROSECONDS = 'microseconds'
